
/**
 * The checkpoints completed in the current lap.
 */
array<CpData@> nottaken = {};

/**
 * All the checkpoints in the current map
 */
array<CpData@> cps = {};

enum WaypointType {
    Start = 0,
    Finish = 1,
    Checkpoint = 2,
    None = 3,
    Multilap = 4,
    Dispenser = 5
}


class CpData {

    uint Value; // unique value to identify each checkpoint, from cp.MarkerId.Value
    string Tag; // cp.tag (start, finish, checkpoint, linkedcheckpoint)
    uint Order; // group (only applies to linked checkpoint)
    vec3 coords; // position of checkpoint in world
#if MP4
    nat3 worldcoords;
    bool istall = false;
#endif
    vec4 outerColor; // outer color for checkpoint dot
    vec4 innerColor; // inner color for cp dot
    vec4 textColor;// = vec4(1, 0, 1, 1); // outerColor for text on cps (used to show order)
    string text; // text that will be shown over cp (cp order (group for linked cps))
    WaypointType waypointType;

    CpData(CGameScriptMapLandmark@ cp) {
#if TMNEXT
        if (maptype == "FlagRushArena" && (cp.Tag.StartsWith("Unused")) || cp.Tag == 'FlagResetTrigger') {return;}
        this.Value = cp.MarkerId.Value;
#elif MP4
        this.Value = 0;
        this.worldcoords = PosToCoord(cp.Position);
#endif
        this.Tag = cp.Tag;
        this.Order = cp.Order;
        this.coords = cp.Position;
        this.text = "";
        if (cp.Waypoint is null) {
            waypointType = WaypointType::Start;
        } else if (cp.Waypoint.IsMultiLap) {
            waypointType = WaypointType::Multilap;
        } else if (cp.Waypoint.IsFinish) {
            waypointType = WaypointType::Finish;
        } else {
            waypointType = WaypointType::Checkpoint;
        }
        if (cp.Order != 0 && !isOption(cp.Order, Options::xdd)) {
            this.text = tostring(cp.Order);
        }
        this.setColor();
    }
    CpData(CGameCtnBlock@ block) {
        CGameWaypointSpecialProperty@ wp = block.WaypointSpecialProperty;
        if (maptype == "FlagRushArena" && (wp.Tag.StartsWith("Unused")) || wp.Tag == 'FlagResetTrigger') {return;}
        this.Value = 0;
        this.Tag = wp.Tag;
        this.Order = wp.Order;
        this.coords = GetCtnBlockMidpoint(block);
#if MP4
        this.worldcoords = block.Coord;
        if (block.BlockModel.Name.EndsWith('CheckpointRingV') || block.BlockModel.Name.EndsWith('CheckpointRing2x1V')) {
            this.istall = true;
        }
#endif
        this.text = "";
        if (block.BlockInfo.WaypointType == CGameCtnBlockInfo::EWayPointType::Start) {
            waypointType = WaypointType::Start;
        } else if (block.BlockInfo.WaypointType == CGameCtnBlockInfo::EWayPointType::Finish) {
            waypointType = WaypointType::Finish;
        } else if (block.BlockInfo.WaypointType == CGameCtnBlockInfo::EWayPointType::Checkpoint) {
            waypointType = WaypointType::Checkpoint;
        } else if (block.BlockInfo.WaypointType == CGameCtnBlockInfo::EWayPointType::StartFinish) {
            waypointType = WaypointType::Multilap;
        } else if (block.BlockInfo.WaypointType == CGameCtnBlockInfo::EWayPointType::Dispenser) {
            waypointType = WaypointType::Dispenser;
        } else {
            waypointType = WaypointType::None;
        }
        if (wp.Order != 0 && !isOption(wp.Order,Options::xdd)) {
            this.text = tostring(wp.Order);
        }
        this.setColor();
    }
    CpData(CGameCtnAnchoredObject@ item) {
        CGameWaypointSpecialProperty@ wp = item.WaypointSpecialProperty;
        if (maptype == "FlagRushArena" && (wp.Tag.StartsWith("Unused")) || wp.Tag == 'FlagResetTrigger') {return;}
        this.Value = 0;
        this.Tag = wp.Tag;
        this.Order = wp.Order;
        this.coords = item.AbsolutePositionInMap;
#if MP4
        this.worldcoords = PosToCoord(item.AbsolutePositionInMap);
#endif
        this.text = "";
        if (item.ItemModel.WaypointType == CGameItemModel::EnumWaypointType::Start) {
            waypointType = WaypointType::Start;
        } else if (item.ItemModel.WaypointType == CGameItemModel::EnumWaypointType::Finish) {
            waypointType = WaypointType::Finish;
        } else if (item.ItemModel.WaypointType == CGameItemModel::EnumWaypointType::Checkpoint) {
            waypointType = WaypointType::Checkpoint;
        } else if (item.ItemModel.WaypointType == CGameItemModel::EnumWaypointType::StartFinish) {
            waypointType = WaypointType::Multilap;
        } else if (item.ItemModel.WaypointType == CGameItemModel::EnumWaypointType::Dispenser) {
            waypointType = WaypointType::Dispenser;
        } else {
            waypointType = WaypointType::None;
        }
        if (wp.Order != 0 && !isOption(wp.Order,Options::xdd)) {
            this.text = tostring(wp.Order);
        }
        this.setColor();
    }
    void setColor() {
        if (waypointType == WaypointType::Start) {
            // Tag 'Spawn' in normal gamemode
            this.outerColor = StartOuterColor;
            this.innerColor = StartInnerColor;
            this.textColor = StartTextColor;
        } else if (waypointType == WaypointType::Finish) {
            // Tag 'Goal' in normal gamemode
            this.outerColor = FinishOuterColor;
            this.innerColor = FinishInnerColor;
            this.textColor = FinishTextColor;
        } else if (waypointType == WaypointType::Multilap) {
            // Tag 'StartFinish' in normal gamemode
            this.outerColor = MultilapOuterColor;
            this.innerColor = MultilapInnerColor;
            this.textColor = MultilapTextColor;
        } else if (waypointType == WaypointType::Checkpoint) {
            if (this.Tag == 'LinkedCheckpoint' || this.Tag == 'FlagSpawn') {
                this.outerColor = LinkedCheckpointOuterColor;
                this.innerColor = LinkedCheckpointInnerColor;
                this.textColor = LinkedCheckpointTextColor;
            } else if (this.Tag == 'Checkpoint' || this.Tag == 'DefaultFlagSpawn') {
                this.outerColor = CheckpointOuterColor;
                this.innerColor = CheckpointInnerColor;
                this.textColor = CheckpointTextColor;
            } else {
                if (this.Tag != "Goal") {
                    print('Unknown Checkpoint Tag: ' + this.Tag);
                }
                this.outerColor = OtherCheckpointOuterColor;
                this.innerColor = OtherCheckpointInnerColor;
                this.textColor = OtherCheckpointTextColor;
            }
        } else if (waypointType == WaypointType::Dispenser) {
            // Tag 'StartFinish' in normal gamemode
            this.outerColor = DispenserOuterColor;
            this.innerColor = DispenserInnerColor;
            this.textColor = DispenserTextColor;
        } else if (waypointType == WaypointType::None) {
            // Tag 'StartFinish' in normal gamemode
            this.outerColor = NoneOuterColor;
            this.innerColor = NoneInnerColor;
            this.textColor = NoneTextColor;
        }
    }
    /*void NoStandingRespawn() { // currently I can't tell this, but if I find a way I will color it differently
        this.outerColor = vec4(0, 0, 0.8, 0.9);
    }*/
    bool opEquals(CpData@ cp) {
        return this is cp;
    }
}

