
void Update_FromMapLandmarks() {
    CGameCtnChallenge@ rootMap = cast<CGameCtnChallenge>(GetApp().RootMap);
    CSmArenaClient@ playground = cast<CSmArenaClient>(GetApp().CurrentPlayground);

    MwFastBuffer<CGameScriptMapLandmark@> landmarks = playground.Arena.MapLandmarks;

    curMapId = rootMap.IdName;
    maptype = rootMap.MapType;
    string comment = rootMap.Comments;

    int start = -1;
    array<int> LinkOrders = {};
    for(uint i = 0; i < landmarks.Length; i++) {
        if ((landmarks[i].Waypoint is null)) {
            start = i;
        } else {
            if (landmarks[i].Waypoint.IsMultiLap) {
                StartOrder = landmarks[i].Order; // will be overwritten if any start block
                
            } else if (landmarks[i].Waypoint.IsFinish) {
                // do nothing
            } else {
            // if (landmarks[i].Tag == 'Checkpoint' || landmarks[i].Tag == 'LinkedCheckpoint') {
                // must be checkpoint, but converted tmnf maps will have tag goal
                if (landmarks[i].Order == 0) {
                    unordered++;
                } else {
                    // if multiple CPs with same order, this order should occur that many times in this list
                    CPorders.InsertLast(landmarks[i].Order);
                    unordered++;
                }
                if (landmarks[i].Tag == 'LinkedCheckpoint') {
                    if (LinkOrders.Find(landmarks[i].Order) == -1) {
                        LinkOrders.InsertLast(landmarks[i].Order);
                        CPcount++;
                    }
                } else {
                    CPcount++;
                }
            }
        }
#if TMNEXT
        if (i | 0x03000000 != landmarks[i].MarkerId.Value) {
            // UI::ShowNotification("Unclaimed Checkpoints Indicator", "Checkpoint MarkerId doesn't match its Landmarks index", vec4(0.5,0,0,0.5));
            print('Checkpoint MarkerId doesn\'t match its landmarks index');
        }
#endif
        cps.InsertLast(CpData(landmarks[i]));
    }
    CPorders.SortAsc();
    if (start != -1){
        StartOrder = landmarks[start].Order;
    }
    if (comment.ToLower().Contains('/uci order') || comment.ToLower().Contains('/uci lowest')) {
        if (StartOrder < 65535) {StartOrder = 65535;}
        StartOrder = ((StartOrder - 65535) | Options::Lowest) + 65535;
    }
    if (comment.ToLower().Contains('/uci hide')) {
        if (StartOrder < 65535) {StartOrder = 65535;}
        StartOrder = ((StartOrder - 65535) | Options::Hide) + 65535;
    }
    if (comment.ToLower().Contains('/uci xdd')) {
        if (StartOrder < 65535) {StartOrder = 65535;}
        StartOrder = ((StartOrder - 65535) | Options::xdd) + 65535;
    }
    if (comment.ToLower().Contains('/uci always')) {
        if (StartOrder < 65535) {StartOrder = 65535;}
        StartOrder = ((StartOrder - 65535) | Options::Always) + 65535;
    }
    partialReset();
    setCurrentMapSettings();
}

#if TMNEXT
void Update_InMap() {
    CGameCtnChallenge@ rootMap = cast<CGameCtnChallenge>(GetApp().RootMap);
    CSmArenaClient@ playground = cast<CSmArenaClient>(GetApp().CurrentPlayground);

    MwFastBuffer<CGameScriptMapLandmark@> landmarks = playground.Arena.MapLandmarks;

    if(!inMap && (curMapId != rootMap.IdName || GetApp().Editor !is null)) {
        // keep the previously-determined CP data, unless in the map editor

        reset(); // won't be called before if in editor

        inMap = true;

        Update_FromMapLandmarks();

        if (allowUnsafeCheckpointsList && !temporaryCheckpointsListDisable && isNewExeVersion && !ignoreExeVersion) {
            showNewExeWindow = true;
        }
    }

    if (playground.Map is null
        || playground.GameTerminals.Length <= 0
        || playground.GameTerminals[0].UISequence_Current != CGamePlaygroundUIConfig::EUISequence::Playing
        || cast<CSmPlayer>(playground.GameTerminals[0].GUIPlayer) is null
        || cps.Length == 0) {
        partialReset();
        return;
    }

    CSmPlayer@ player = cast<CSmPlayer>(playground.GameTerminals[0].GUIPlayer);

    // this might be relevant when spectating players in multiplayer (you can't see their CPs)
    if(player.CurrentLaunchedRespawnLandmarkIndex == uint(-1)) {
        partialReset();
        return;
    }

    if (!inGame) {
        inGame = true;
        preCPIdx = uint(-2); // this will never match current CP, so will always re-check CPs
    }

    if(preCPIdx != player.CurrentLaunchedRespawnLandmarkIndex && landmarks.Length > player.CurrentLaunchedRespawnLandmarkIndex) {
        // taken new CP
        uint curCPIdx = player.CurrentLaunchedRespawnLandmarkIndex;

        if (maptype == "FlagRushArena") {
            if (landmarks[curCPIdx].Tag == "SpawnTeam1") {
                StartOrder = 1;
            } else if (landmarks[curCPIdx].Tag == "SpawnTeam2") {
                StartOrder = 2;
            }
        } else {
            if (landmarks[curCPIdx].Waypoint is null || landmarks[curCPIdx].Waypoint.IsFinish || landmarks[curCPIdx].Waypoint.IsMultiLap) {
                // if null, it's a start block. if the other flags, it's either a multilap or a finish. In all such cases, reset the nottaken cps to all cps.
                if (maptype == "TrackMania\\TM_Royal") {
                    if (landmarks.Length > player.CurrentLaunchedRespawnLandmarkIndex) { // not editor-spawn without CP
                        StartOrder = landmarks[curCPIdx].Order; // it might be a finish but that doesn't matter, order should still be the same
                    }
                }

                LapReset();

            } else {
                // Checkpoint
                array<uint> toremove = {};
                if (allowUnsafeCheckpointsList && !temporaryCheckpointsListDisable && (ignoreExeVersion || !isNewExeVersion)) {
                    string playername = player.ScriptAPI.Name;

                    // Based on code from XertroV, who found the taken checkpoints list
                    CGameCtnReplayRecord@ currRR = cast<CGameCtnReplayRecord>(Dev::GetOffsetNod(playground.Arena, Reflection::TypeOf(playground.Arena).GetMember("Rules").Offset + (0x9F0 - 0x938)));
                    if (currRR !is null && currRR.Ghosts.Length > 0) {
                        uint16 bufOffset = Reflection::GetType("CGameCtnGhost").GetMember("NbRespawns").Offset + 0x8;
                        uint16 bufLenO = bufOffset + 0x8;
                        uint ghostindex = uint(-1);
                        for (uint k = currRR.Ghosts.Length-1; k > 0; k--) {
                            // print('' + k + ': ' + currRR.Ghosts[k].GhostNickname + ' (' + Dev::GetOffsetUint32(currRR.Ghosts[k], bufLenO) + ')');
                            if (currRR.Ghosts[k].GhostNickname == playername && ghostindex == uint(-1)) {
                                ghostindex = k;
                            }
                        }
                        if (ghostindex == uint(-1)) {
                            print('Player ghost not found: ' + playername);
                            // for (uint k = 0; k < currRR.Ghosts.Length; k++) {
                            //     print('' + string(currRR.Ghosts[k].GhostNickname) + ' ' + string(currRR.Ghosts[k].GhostNickname) == playername);
                            // }
                            ghostindex = 0;
                        }
                        CGameCtnGhost@ g = currRR.Ghosts[ghostindex];
                        uint64 cpBufPtr = Dev::GetOffsetUint64(g, bufOffset);
                        uint cpBufLen = Dev::GetOffsetUint32(g, bufLenO);
                        // values (ids) of checkpoints taken this frame
                        array<uint> values = {};
                        // orders of linked Checkpoints taken this frame
                        array<uint> linkedCPOrders = {};

                        // print('cpBufLen: ' + cpBufLen);
                        for (uint j = cpBufLen-1; j != uint(-1) && j >= 0; j--) {
                            // print(j);
                            // (id, time taken) where id is just landmarks[i] not full MarkerId
                            nat2 cpIdTime = Dev::ReadNat2(cpBufPtr + j * 0x8);
                            uint bufferCPIdx = cpIdTime.x;
                            uint bufferCPValue = cpIdTime.x | 0x03000000;
                            if (bufferCPIdx == preCPIdx) {break;}
                            // print(Id);
                            if (j == cpBufLen-1 && bufferCPIdx != curCPIdx) {
                                print('YEK latest CP in buffer is different to moving respawn: ' + bufferCPIdx + ', ' + curCPIdx);
                                print('Consider disabling accessing CPs list memory location');
                                UI::ShowNotification("Unclaimed Checkpoints Indicator", "Issues detected, consider disabling taken CP list memory access", vec4(0.5,0,0,0.5));
                            } else {
                                if (isNewExeVersion) {
                                    savedExeVersion = GetApp().SystemPlatform.ExeVersion;
                                    isNewExeVersion = false;
                                    Meta::SaveSettings();
                                    ignoreExeVersion = false;
                                    showNewExeWindow = false;
                                    temporaryCheckpointsListDisable = false;
                                }
                            }
                            values.InsertLast(landmarks[bufferCPIdx].MarkerId.Value);
                            if (landmarks[bufferCPIdx].Tag == 'LinkedCheckpoint') {
                                linkedCPOrders.InsertLast(landmarks[bufferCPIdx].Order);
                            }
                        }
                        for (uint i = 0; i < nottaken.Length; i++){
                            if (values.Find(nottaken[i].Value) != -1) {
                                toremove.InsertLast(i);
                                takenCPcount++;
                            } else if (nottaken[i].Tag == 'LinkedCheckpoint' && linkedCPOrders.Find(nottaken[i].Order) != -1) {
                                toremove.InsertLast(i);
                            } else if (nottaken[i].waypointType == WaypointType::Start) {
                                toremove.InsertLast(i);
                            }
                        }
                    }
                } else {
                    uint value = landmarks[curCPIdx].MarkerId.Value;
                    for (uint i = 0; i < nottaken.Length; i++){
                        if (nottaken[i].Value == value) {
                            toremove.InsertLast(i);
                        } else if (landmarks[curCPIdx].Tag == 'LinkedCheckpoint' && nottaken[i].Tag == 'LinkedCheckpoint' && nottaken[i].Order == landmarks[curCPIdx].Order) {
                            toremove.InsertLast(i);
                        }
                    }
                }
                toremove.SortDesc(); // toremove is a list of checkpoint indexes to remove from nottaken, in order from last to first
                for (uint i = 0; i < toremove.Length; i++) {
                    if (nottaken[toremove[i]].Order != 0) {
                        uint j = 0;
                        for (; j < NextCPorders.Length; j++) {
                            if (NextCPorders[j] == nottaken[toremove[i]].Order) {break;}
                        }
                        if (j < NextCPorders.Length) {
                            NextCPorders.RemoveAt(j);}
                        else {
                            print('cp orders error (I think I\'ve fixed this)');
                        }
                    }
                    if (toremove[i] >= nottaken.Length) {
                        print('trying to remove CP after end of not taken');
                        print(toremove[i] + " " + toremove.Length + " " + nottaken.Length);
                    } else {
                        nottaken.RemoveAt(toremove[i]);
                    }
                }
            }
        }
        preCPIdx = curCPIdx;
    }
}
#elif MP4
// int checkpointnumber = -1;
void Update_InMap() {
    CGameCtnChallenge@ rootMap = cast<CGameCtnChallenge>(GetApp().RootMap);

    if (!inMap && (curMapId != rootMap.IdName || GetApp().Editor !is null)) {
        // keep the previously-determined CP data, unless in the map editor

        reset(); // won't be called before if in editor

        inMap = true;

        Update_FromRootmap();
    }

    CTrackManiaRace@ playground = cast<CTrackManiaRace>(GetApp().CurrentPlayground);

    if (playground.GameTerminals.Length <= 0
        || cast<CTrackManiaPlayer>(playground.GameTerminals[0].GUIPlayer) is null) {
        partialReset();
        return;
    }

    CTrackManiaPlayer@ player = cast<CTrackManiaPlayer>(playground.GameTerminals[0].GUIPlayer);

    // if (player.ScriptAPI.CurLap.Checkpoints.Length > 0) {
    //     if (player.ScriptAPI.CurLap.Checkpoints[player.ScriptAPI.CurLap.Checkpoints.Length - 1] != checkpointnumber) {
    //         // print(player.ScriptAPI.CurLap.Checkpoints[player.ScriptAPI.CurLap.Checkpoints.Length - 1]);
    //         checkpointnumber = player.ScriptAPI.CurLap.Checkpoints[player.ScriptAPI.CurLap.Checkpoints.Length - 1];
    //     }
    // }
    takenCPcount = player.ScriptAPI.CurLap.Checkpoints.Length;
    // print(player.ScriptAPI.CurTriggerIndex);

    if (!inGame) {
        inGame = true;
        preCPTime = -2; // this will never match current CP, so will always re-check CPs
        previousSpawnLoc = nat3(uint(-1),uint(-1),uint(-1));
    }

    if(player.RaceState == CTrackManiaPlayer::ERaceState::Running && ((player.ScriptAPI.CurLap.Checkpoints.Length != 0 && preCPTime != player.ScriptAPI.CurLap.Checkpoints[player.ScriptAPI.CurLap.Checkpoints.Length - 1]) || (player.ScriptAPI.CurLap.Checkpoints.Length == 0 && preCPTime != -1))) {
        // taken new CP

        if (player.ScriptAPI.CurLap.Checkpoints.Length == 0) {
            preCPTime = -1;
        } else {
            preCPTime = player.ScriptAPI.CurLap.Checkpoints[player.ScriptAPI.CurLap.Checkpoints.Length - 1];
        }

        nat3 takenCPpos = PosToCoord(vec3(player.SpawnLoc.tx, player.SpawnLoc.ty, player.SpawnLoc.tz));
        // print('Taken at: ' + takenCPpos.ToString());
        bool noCP = true;
        // If its a ring CP, then the taken CP coords don't change
        if (takenCPpos != previousSpawnLoc) {
            takenCPpos = PosToCoord(player.Position);
        } else {
            // intentionally not changing this if its a ring CP so multiple rings in a row
            previousSpawnLoc = takenCPpos;
        }
        for (uint CPidx = 0; CPidx < cps.Length; CPidx++) {
            // print("checking at: " + cps[CPidx].worldcoords.ToString());
            if (cps[CPidx].worldcoords == takenCPpos || cps[CPidx].worldcoords == takenCPpos - nat3(0, 1, 0) || (cps[CPidx].istall && (cps[CPidx].worldcoords == takenCPpos - nat3(0, 2, 0) || cps[CPidx].worldcoords == takenCPpos - nat3(0, 3, 0))))  {
                // this is the CP we've taken (unless they're mix blocked)
                noCP = false;
                CpData@ CP = cps[CPidx];
                if (CP.waypointType == WaypointType::Start || CP.waypointType == WaypointType::Finish || CP.waypointType == WaypointType::Multilap) {
                    LapReset();
                    currentLap = player.CurrentNbLaps;
                } else {
                    // Checkpoint
                    array<uint> toremove = {};
                    uint value = CP.Value;
                    for (uint i = 0; i < nottaken.Length; i++){
                        if (nottaken[i] == CP) {
                            toremove.InsertLast(i);
                        } else if (CP.Tag == 'LinkedCheckpoint' && nottaken[i].Tag == 'LinkedCheckpoint' && nottaken[i].Order == CP.Order) {
                            toremove.InsertLast(i);
                        }
                    }
                    toremove.SortDesc(); // toremove is a list of checkpoint indexes to remove from nottaken, in order from last to first
                    for (uint i = 0; i < toremove.Length; i++) {
                        if (nottaken[toremove[i]].Order != 0) {
                            uint j = 0;
                            for (; j < NextCPorders.Length; j++) {
                                if (NextCPorders[j] == nottaken[toremove[i]].Order) {break;}
                            }
                            if (j < NextCPorders.Length) {
                                NextCPorders.RemoveAt(j);}
                            else {
                                print('cp orders error (I think I\'ve fixed this)');
                            }
                        }
                        if (toremove[i] >= nottaken.Length) {
                            print('trying to remove CP after end of not taken');
                            print(toremove[i] + " " + toremove.Length + " " + nottaken.Length);
                        } else {
                            nottaken.RemoveAt(toremove[i]);
                        }
                    }
                }
            }
        }
        if (noCP) {
            trace('respawn location not a CP position');
            if (preCPTime == -1 || preLapCount != player.CurrentNbLaps) {
                // start, new lap or finish (so reset CPs)

                LapReset();
                currentLap = player.CurrentNbLaps;
            }
        }
    }
}

void Update_InSMMap() {
    CGameCtnChallenge@ rootMap = cast<CGameCtnChallenge>(GetApp().RootMap);

    if (!inMap && (curMapId != rootMap.IdName || GetApp().Editor !is null)) {
        // keep the previously-determined CP data, unless in the map editor

        reset(); // won't be called before if in editor

        inMap = true;

        // Update_FromMapLandmarks();
        Update_FromRootmap();
    }

    CSmArenaClient@ playground = cast<CSmArenaClient>(GetApp().CurrentPlayground);

    if (playground.GameTerminals.Length <= 0
        || cast<CSmPlayer>(playground.GameTerminals[0].GUIPlayer) is null) {
        partialReset();
        return;
    }

    // CSmPlayer@ player = cast<CSmPlayer>(playground.GameTerminals[0].GUIPlayer);

    takenCPcount = 0;

    if (!inGame) {
        inGame = true;
        preCPTime = -2; // this will never match current CP, so will always re-check CPs
        previousSpawnLoc = nat3(uint(-1),uint(-1),uint(-1));
    }
}

#endif
