/*
 * by ArEyeses
 * (some code copied with permission from CP counter by Phlarx)
 */

/* there are some parts that are commented out because I can't find a way to make them work
 * - Type of checkpoint block e.g. ring (I can find this, but can't tell which block/item is which checkpoint)
 * - rotation of block when in freemode (also needs above point to be useful)
 */

[Setting category="General" name="Show when driving"]
bool show = true;

[Setting category="General" name="Track CPs when hidden" description="If 'use checkpoint list' is disabled, then this allows you to show the plugin to find the last cp"]
bool run = true;

[Setting category="General" name="Show in Editor" description="Show in any mode where you are not driving, such as map editor or replay editor"]
bool showEditor = true;

#if TMNEXT
[Setting category="General" name="Use checkpoint list" description="May be unsafe, its directly accessing memory location"]
bool allowUnsafeCheckpointsList = true;
#endif

[Setting category="General" name="Dot size" min=2 max=64]
int dotSize = 7;

[Setting category="General" name="Show numbers above dot"]
bool showNumberUp = true;

[Setting category="General" name="Show when UI hidden"]
bool showUIHidden = true;

[Setting category="General" name="Show when pause menu open (*pause menu detection unreliable)"]
bool showPause = true;

[Setting category="General" name="Smaller dots with distance" description="Further checkpoints have smaller dots. Will never become too small to see"]
bool showDistance = true;

[Setting category="General" name="Use gradient colors"]
bool useGradientFill = false;

[Setting category="General" name="Single-color dots (small rendering speed increase)"]
bool lowDrawQuality = false;

#if MP4
[Setting category="General" name="Exclude Titlepacks" description="format as e.g. `TMCanyon@nadeo, TMStadium@nadeo`"]
string excludedTitlepacks_Input = '';
string[] excludedTitlepacks = {};
#endif

#if TMNEXT
[Setting hidden]
string savedExeVersion = '2023-11-24_17_34';
[Setting hidden]
bool temporaryCheckpointsListDisable = false;

bool isNewExeVersion = true;
bool ignoreExeVersion = false;
bool showNewExeWindow = false;

bool newExeVersion() {
    return GetApp().SystemPlatform.ExeVersion > savedExeVersion;
}

void updateExeVersionFromSite() {
    trace('requesting known safe exe version from site');
    Net::HttpRequest@ getVer = Net::HttpRequest();
    getVer.Url = "https://openplanet.dev/plugin/unclaimedcheckpointindicator/config/savedexeversion";
    getVer.Start();
    while (!getVer.Finished()) {
        yield();
    }
    if (getVer.ResponseCode() == 200) {
        if (getVer.String().Length > 0) {
            string ver = Json::Parse(getVer.String())[0];
            if (ver > savedExeVersion) {
                savedExeVersion = ver;
                Meta::SaveSettings();
                isNewExeVersion = newExeVersion();
                if (!isNewExeVersion) { // if downloads an updated version but it isn't the current version
                    ignoreExeVersion = false;
                    showNewExeWindow = false;
                    temporaryCheckpointsListDisable = false;
                } else {
                    if (allowUnsafeCheckpointsList && inGame) {
                        reset();
                    }
                }
            }
        }
    }
}
#endif

// using Options enum; this means you can disable the map options later
uint currentMapSettings = 0;

/**
 * If true, then this plugin has detected that we are on a map.
 * If this and inEditor false, none of the other values are valid.
 */
bool inMap = false;
// editor version of inMap
bool inEditor = false;
// if the car is actually driving (not in an intro/finished)
bool inGame = false;

/**
 * The ID of the map whose checkpoints are currently in the arraya.
 */
string curMapId = "";

/**
 * The current respawn location (thus the most recent cp taken).
 * when this changes, it means we've taken a new cp
 */
#if TMNEXT
uint preCPIdx = uint(-2);
#elif MP4
int preCPTime = -2;
uint preLapCount = uint(-1);
nat3 previousSpawnLoc = nat3(uint(-1),uint(-1),uint(-1));
#endif

// the order of the start block. If mapper sets this to __, will only show checkpoints in order
uint StartOrder = 0;

// number of checkpoints with no order
uint unordered = 0;
// orders of checkpoints used in the map (for if they will be displayed in order). Does contain duplicates
array<uint> CPorders = {};
// orders of checkpoints that haven't been taken (only used to filter display)
array<uint> NextCPorders = {};

string maptype = "";

uint64 startTime;

enum Options {
    // only show the checkpoint with lowest remaining order
    Lowest = 0b1,
    // hide all checkpoints (except those with 'Always')
    Hide = 0b10,
    xdd = 0b100,
    Always = 0b1000,
}
bool isOption(uint order, uint option) {
    return (order > 65535 && (order - 65535)&option != 0);
}

void OnSettingsChanged() {
    for (uint i = 0; i < cps.Length; i++) {
        cps[i].setColor();
    }
#if MP4
    excludedTitlepacks = excludedTitlepacks_Input.Split(',');
    for (uint i = 0; i < excludedTitlepacks.Length; i++) {
        excludedTitlepacks[i] = excludedTitlepacks[i].Trim().ToLower();
    }
#endif
#if TMNEXT
    if (!allowUnsafeCheckpointsList) {
        if (showNewExeWindow) {
            showNewExeWindow = false;
        }
        if (temporaryCheckpointsListDisable) {
            temporaryCheckpointsListDisable = false;
        }
    } else if (allowUnsafeCheckpointsList && isNewExeVersion && !ignoreExeVersion) {
        showNewExeWindow = true;
    }
#endif
    setCurrentMapSettings();
}

// exit map, resets the current map and checkpoint lists
void reset() {
    curMapId = "";
    nottaken = {};
    cps = {};
    inGame = false;
    inMap = false;
    inEditor = false;
    StartOrder = 0;
    unordered = 0;
    CPorders = {};
    NextCPorders = {};
    maptype = "";
    currentMapSettings = 0;
    CPcount = 0;
    takenCPcount = 0;
    currentLap = 0;

}
// not driving but in map e.g. finished mode
void partialReset() {
    inGame = false;
    if (maptype == "TrackMania\\TM_Royal") {
        StartOrder = 0; // multiple starts, so will be set once you've spawned from one of them
        NextCPorders = CPorders;
    } else if (maptype == "FlagRushArena") {
        StartOrder = 0; // multiple starts, so will be set once you've spawned from one of them
    } else {
        NextCPorders = CPorders;
    }
    nottaken = cps;
    takenCPcount = 0;
    currentLap = 0;
}
// reset when dnf, new lap or restart while still in driving mode
void LapReset() {
    nottaken = {};
    nottaken.Reserve(cps.Length - 1); // if multilap, this will be extended to 1 more
    for (uint i = 0; i < cps.Length; i++) {
        if (cps[i].waypointType != WaypointType::Start) {
            nottaken.InsertLast(cps[i]);
        }
    }
    takenCPcount = 0;
    NextCPorders = CPorders;
}

void Main() {
#if MP4
    excludedTitlepacks = excludedTitlepacks_Input.Split(',');
    for (uint i = 0; i < excludedTitlepacks.Length; i++) {
        excludedTitlepacks[i] = excludedTitlepacks[i].Trim().ToLower();
    }
#endif
#if TMNEXT
    isNewExeVersion = newExeVersion();
    if (isNewExeVersion) {
        startnew(updateExeVersionFromSite);
    } else if (temporaryCheckpointsListDisable) {
        temporaryCheckpointsListDisable = false;
    }
#endif
    while (true) {
        yield();
        startTime = Time::get_Now();

        if (!run && !show && !showEditor) {continue;}
        
        if (GetApp().RootMap is null) {
            reset();
            continue;
        }

#if TMNEXT
        if (run || show) {
            CSmArenaClient@ playground = cast<CSmArenaClient>(GetApp().CurrentPlayground);
            if(playground !is null
                && playground.Arena !is null) {
                Update_InMap();
                continue;
            }
        }
#elif MP4
        if (run || show) {
            CTrackManiaRace@ playground = cast<CTrackManiaRace>(GetApp().CurrentPlayground);
            if(playground !is null &&
                playground.GameTerminals.Length > 0) {
                Update_InMap();
                continue;
            }

            CSmArenaClient@ playgroundSm = cast<CSmArenaClient>(GetApp().CurrentPlayground);
            if(playgroundSm !is null &&
                playgroundSm.GameTerminals.Length > 0) {
                Update_InSMMap();
                continue;
            }
        }
#endif

        // playground is null when loading in to map, so need to check editor exists (could be map or replay)
        if (GetApp().Editor is null) {
            reset();
            continue;
        } else if (showEditor) {
            Update_InEditor();
        }
        // continue;
    }
}

void setCurrentMapSettings() {
    currentMapSettings = 0;
    CGameCtnChallenge@ rootMap = cast<CGameCtnChallenge>(GetApp().RootMap);
    if (rootMap is null) {return;}
    // StartOrder variable will be changed if there are relevant map comments

    if (isOption(StartOrder, Options::Lowest)) {
        currentMapSettings |= Options::Lowest;
    }
    if (isOption(StartOrder, Options::Hide)) {
        currentMapSettings |= Options::Hide;
    }
    if (isOption(StartOrder, Options::xdd)) {
        currentMapSettings |= Options::xdd;
    }
    if (isOption(StartOrder, Options::Always)) {
        currentMapSettings |= Options::Always;
    }
#if MP4
    CGameManiaTitle@ titlepack = cast<CGameManiaTitle>(cast<CGameManiaPlanet>(GetApp()).LoadedManiaTitle);
    if (excludedTitlepacks.Length > 0 && excludedTitlepacks.Find(titlepack.TitleId.ToLower()) != -1) {
        currentMapSettings |= Options::Hide;
    }
#endif
}
